@@ -37,7 +37,7 @@ class Session:
         """
 
         # verify auth endpoint's server certificate if proxy is type SOCKS or None
-        if proxy == None:
+        if proxy is None:
             self.proxies = dict()
             self.verify = True
         else:
@@ -358,7 +358,7 @@ class Organization:
                 "findByNetworkGroupId": network_group_id
             }
             response = requests.get(
-                self.session.audience+'/core/v2/networks',
+                self.session.audience+'core/v2/networks',
                 proxies=self.session.proxies,
                 verify=self.session.verify,
                 headers=headers,
@@ -495,8 +495,9 @@ class NetworkGroup:
             # data centers returns a list of dicts (data center objects)
             headers = { "authorization": "Bearer " + self.session.token }
             params = {
-                "hostType": "NC",
-                "provider": "AWS"
+                # "productVersion": self.product_version,
+                # "hostType": "NC",
+                # "provider": "AWS"
             }
             response = requests.get(
                 self.session.audience+'core/v2/data-centers',
@@ -511,7 +512,8 @@ class NetworkGroup:
 
         if response_code == requests.status_codes.codes.OK: # HTTP 200
             try:
-                data_centers = json.loads(response.text)['_embedded']['dataCenters']
+                all_data_centers = json.loads(response.text)['_embedded']['dataCenters']
+                aws_data_centers = [dc for dc in all_data_centers if dc['provider'] == "AWS"]
             except ValueError as e:
                 eprint('ERROR getting data centers')
                 raise(e)
@@ -524,57 +526,28 @@ class NetworkGroup:
                 )
             )
 
-        return(data_centers)
+        return(aws_data_centers)
 
-    def get_data_center_by_location(self, location):
-        """return one data center object
-        :param location: required single location to fetch
-        """
-        try:
-            # data centers returns a list of dicts (data center objects)
-            headers = { "authorization": "Bearer " + self.session.token }
-            params = { "locationCode": location }
-            response = requests.get(
-                self.session.audience+'rest/v1/dataCenters',
-                proxies=self.session.proxies,
-                verify=self.session.verify,
-                headers=headers,
-                params=params
-            )
-            response_code = response.status_code
-        except:
-            raise
-
-        if response_code == requests.status_codes.codes.OK: # HTTP 200
-            try:
-                data_center = json.loads(response.text)
-            except ValueError as e:
-                eprint('ERROR getting data center')
-                raise(e)
-        else:
-            raise Exception(
-                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
-                    requests.status_codes._codes[response_code][0].upper(),
-                    response_code,
-                    response.text
-                )
-            )
-
-        return(data_center)
-
-    def create_network(self, name, network_group_id=None, location="us-east-1", version=None, network_config="small"):
+    def create_network(self, name: str, network_group_id: str=None, location: str="us-east-1", version: str=None, size: str="small"):
         """
         create a network with
         :param name: required network name
         :param network_group: optional Network Group ID
         :param location: optional data center region name in which to create
-        :param version: optional product version string like 7.2.0-1234567
-        :param network_config: optional network configuration metadata name e.g. "medium"
+        :param version: optional product version string like 7.3.17
+        :param size: optional network configuration metadata name from /core/v2/network-configs e.g. "medium"
         """
+        
+        if not size in self.network_config_metadata_by_name.keys():
+            raise Exception("ERROR: unexpected Network size '{:s}'. Valid sizes include: {}.".format(size, str(self.network_config_metadata_by_name.keys())))
+
+        if not location in self.nc_data_centers_by_location.keys():
+            raise Exception("ERROR: unexpected Network location '{:s}'. Valid locations include: {}.".format(location, self.nc_data_centers_by_location.keys()))
+
         request = {
             "name": name,
             "locationCode": location,
-            "networkConfigMetadataId": self.network_config_metadata_by_name[network_config]
+            "size": size,
         }
         if network_group_id:
             request["networkGroupId"] = network_group_id
@@ -610,11 +583,8 @@ class NetworkGroup:
                 )
             )
 
-        network_id = json.loads(response.text)['id']
-        # expected value is UUID
-        UUID(network_id, version=4) # validate the returned value is a UUID
-
-        return(network_id)
+        network = json.loads(response.text)
+        return(network)
 
     def delete_network(self, network_id=None, network_name=None):
         """
@@ -624,7 +594,8 @@ class NetworkGroup:
         """
         try:
             if network_id:
-                network_name = [ net['name'] for net in self.networks_by_name if net['id'] == network_id ][0]
+#                import epdb; epdb.serve()
+                network_name = next(name for name, uuid in self.networks_by_name.items() if uuid == network_id)
             elif network_name and network_name in self.networks_by_name.keys():
                 network_id = self.networks_by_name[network_name]
         except:
@@ -652,7 +623,9 @@ class NetworkGroup:
                 )
             )
 
-        return(True)
+        network = json.loads(response.text)
+        return(network)
+
 class Network:
     """describe and use a Network
     """
@@ -678,7 +651,7 @@ class Network:
         self.status = self.describe['status']
         self.product_version = self.describe['productVersion']
         self.owner_identity_id = self.describe['ownerIdentityId']
-        self.network_confi_metadata_id = self.describe['networkConfigMetadataId']
+        self.size = self.describe['size']
         self.o365_breakout_category = self.describe['o365BreakoutCategory']
         self.created_at = self.describe['createdAt']
         self.updated_at = self.describe['updatedAt']
@@ -731,7 +704,7 @@ class Network:
                 "hostType": "ER"
             }
             if provider is not None:
-                if provider in ["AWS", "AZURE", "GCP", "ALICLOUD", "NetFoundry", "OCP"]:
+                if provider in ["AWS", "AZURE", "GCP", "ALICLOUD", "NETFOUNDRY", "OCP"]:
                     params['provider'] = provider
                 else:
                     raise Exception("ERROR: unexpected cloud provider {:s}".format(provider))
@@ -761,7 +734,9 @@ class Network:
                 )
             )
         if location_code:
-            return([dc for dc in data_centers if dc['locationCode'] == location_code])
+            matching_data_centers = [dc for dc in data_centers if dc['locationCode'] == location_code]
+#            import epdb; epdb.serve()
+            return(matching_data_centers)
         else:
             return(data_centers)
 
@@ -815,8 +790,12 @@ class Network:
                 "size": 10,
                 "sort": "name,asc"
             }
+            if type == "services": 
+                params["beta"] = ''
+
             if name is not None:
                 params['name'] = name
+
             response = requests.get(
                 self.session.audience+'core/v2/'+type,
                 proxies=self.session.proxies,
@@ -888,34 +867,109 @@ class Network:
             return(all_pages)
 
     def patch_resource(self,patch):
-        """return a resources
-            :patch: required dictionary with the new properties 
+        """returns a resource
+            :patch: required dictionary with changed properties 
+        """
+
+        headers = {
+            "authorization": "Bearer " + self.session.token 
+        }
+
+        try:
+            before_response = requests.get(
+                patch['_links']['self']['href'],
+                proxies=self.session.proxies,
+                verify=self.session.verify,
+                headers=headers
+            )
+            before_response_code = before_response.status_code
+        except:
+            raise
+
+        if before_response_code in [requests.status_codes.codes.OK]: # HTTP 200
+            try:
+                before_resource = json.loads(before_response.text)
+            except ValueError as e:
+                eprint('ERROR: failed to load {r} object from GET response'.format(r = type))
+                raise(e)
+        else:
+            json_formatted = json.dumps(patch, indent=2)
+            raise Exception(
+                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s} for GET {:s}'.format(
+                    requests.status_codes._codes[before_response_code][0].upper(),
+                    before_response_code,
+                    before_response.text,
+                    json_formatted
+                )
+            )
+        # compare the patch to the discovered, current state, adding new or updated keys to pruned_patch
+        pruned_patch = dict()
+        for k in patch.keys():
+            if k in before_resource.keys() and not before_resource[k] == patch[k]:
+                pruned_patch[k] = patch[k]
+
+        # attempt to update if there's at least one difference between the current resource and the submitted patch
+        if len(pruned_patch.keys()) > 0:
+            try:
+                after_response = requests.patch(
+                    patch['_links']['self']['href'],
+                    proxies=self.session.proxies,
+                    verify=self.session.verify,
+                    headers=headers,
+                    json=pruned_patch
+                )
+                after_response_code = after_response.status_code
+            except:
+                raise
+            if after_response_code in [requests.status_codes.codes.OK, requests.status_codes.codes.ACCEPTED]: # HTTP 202
+                try:
+                    after_resource = json.loads(after_response.text)
+                except ValueError as e:
+                    eprint('ERROR: failed to load {r} object from PATCH response'.format(r = type))
+                    raise(e)
+            else:
+                json_formatted = json.dumps(patch, indent=2)
+                raise Exception(
+                    'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s} for PATCH update {:s}'.format(
+                        requests.status_codes._codes[after_response_code][0].upper(),
+                        after_response_code,
+                        after_response.text,
+                        json_formatted
+                    )
+                )
+            return(after_resource)
+        else:
+            return(before_resource)
+
+    def put_resource(self,put):
+        """returns a resource
+            :put: required dictionary with all properties required by the particular resource's model 
         """
         try:
             headers = {
                 "authorization": "Bearer " + self.session.token 
             }
-            response = requests.patch(
-                patch['_links']['self']['href'],
+            response = requests.put(
+                put['_links']['self']['href'],
                 proxies=self.session.proxies,
                 verify=self.session.verify,
                 headers=headers,
-                json=patch
+                json=put
             )
             response_code = response.status_code
         except:
             raise
 
-        if response_code in [requests.status_codes.codes.OK, requests.status_codes.codes.ACCEPTED]: # HTTP 200
+        if response_code in [requests.status_codes.codes.OK, requests.status_codes.codes.ACCEPTED]: # HTTP 202
             try:
                 resource = json.loads(response.text)
             except ValueError as e:
-                eprint('ERROR: failed to load {r} object from PATCH response'.format(r = type))
+                eprint('ERROR: failed to load {r} object from PUT response'.format(r = type))
                 raise(e)
         else:
-            json_formatted = json.dumps(patch, indent=2)
+            json_formatted = json.dumps(put, indent=2)
             raise Exception(
-                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s} for patch {:s}'.format(
+                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s} for PUT update {:s}'.format(
                     requests.status_codes._codes[response_code][0].upper(),
                     response_code,
                     response.text,
@@ -1059,10 +1113,14 @@ class Network:
 
         return(policy)
 
-    def create_service(self, name: str, client_host_name: str, client_port_range: int, server_host_name: str, 
-        server_port_range: int, server_protocol: str="TCP", attributes: list=[], edge_router_attributes: list=[], 
+    def create_service(self, name: str, client_host_name: str, client_port_range: int, server_host_name: str=None, 
+        server_port_range: int=None, server_protocol: str="tcp", attributes: list=[], edge_router_attributes: list=["#all"], 
         egress_router_id: str=None, endpoints: list=[], encryption_required: bool=True):
-        """create a Service
+        """create a Service to be accessed by Tunneler Endpoints
+        There are three types of servers that may be published with this method: SDK, Tunneler, or Router. 
+        If server details are absent then the type is inferred to be SDK (Service is hosted by a Ziti SDK,
+        not a Tunneler or Router). If server details are present then the Service is either hosted by a
+        Tunneler or Router, depending on which value is present i.e. Tunneler Endpoint or Edge Router. 
         """
         try:
             headers = { 
@@ -1074,33 +1132,96 @@ class Network:
             body = {
                 "networkId": self.id,
                 "name": name,
-                "attributes": attributes,
-                "clientHostName": client_host_name,
-                "clientPortRange": client_port_range,
-                "serverHostName": server_host_name,
-                "serverPortRange": server_port_range,
-                "serverProtocol": server_protocol,
-                "encryptionRequired": encryption_required
+                "encryptionRequired": encryption_required,
+                "model": {
+                    "clientIngress" : {
+                        "host": client_host_name, 
+                        "port": client_port_range,
+                    },
+                    "edgeRouterAttributes" : edge_router_attributes
+                },
+                "attributes" : attributes,
             }
             # resolve exit hosting params
-            if egress_router_id and endpoints:
-                raise Exception("ERROR: specify only one of egress_router_id or endpoints to host the exit for this Service")
-            elif endpoints:
-                body['endpoints'] = endpoints
-            elif egress_router_id:
-                body['egressRouterId'] = egress_router_id
+            if not server_host_name:
+                body['modelType'] = "TunnelerToSdk"
+                if server_port_range:
+                    eprint("WARN: ignoring unexpected server details for SDK-hosted Service")
+            else:
+                server_egress = {
+                    "protocol": server_protocol.lower(),
+                    "host": server_host_name,
+                    "port": server_port_range
+                }            
+                if endpoints and not egress_router_id:
+                    body['modelType'] = "TunnelerToEndpoint"
+                    # parse out the elements in the list of endpoints as one of #attribute, UUID, or resolvable Endoint name
+                    bind_endpoints = list()
+                    for endpoint in endpoints:
+                        if endpoint[0:1] == '#':
+                            bind_endpoints += [endpoint]
+                        else:
+                            # strip leading @ if present and re-add later after verifying the named Endpoint exists
+                            if endpoint[0:1] == '@':
+                                endpoint = endpoint[1:]
 
+                            # if UUIDv4 then resolve to name, else verify the named Endpoint exists 
+                            try:
+                                UUID(endpoint, version=4) # assigned below under "else" if already a UUID
+                            except ValueError:
+                                # else assume is a name and resolve to ID
+                                try: 
+                                    name_lookup = self.get_resources(type="endpoints",name=endpoint)[0]
+                                    endpoint_name = name_lookup['name']
+                                except Exception as e:
+                                    raise Exception('ERROR: Failed to find exactly one hosting Endpoint named "{}". Caught exception: {}'.format(endpoint, e))
+                                # append to list after successfully resolving name to ID
+                                else: bind_endpoints += ['@'+endpoint_name] 
+                            else:
+                                try:
+                                    name_lookup = self.get_resources(type="endpoints",id=endpoint)
+                                    endpoint_name = name_lookup['name']
+                                except Exception as e:
+                                    raise Exception('ERROR: Failed to find exactly one hosting Endpoint with ID "{}". Caught exception: {}'.format(endpoint, e))
+                                else: bind_endpoints += ['@'+endpoint_name] 
+                    body['model']['bindEndpointAttributes'] = bind_endpoints
+                    body['model']['serverEgress'] = server_egress
+
+                elif egress_router_id and not endpoints:
+                    body['modelType'] = "TunnelerToEdgeRouter"
+                    # check if UUIDv4
+                    try: UUID(egress_router_id, version=4)
+                    except ValueError:
+                        # else assume is a name and resolve to ID
+                        try: 
+                            name_lookup = self.get_resources(type="edge-routers",name=egress_router_id)[0]
+                            egress_router_id = name_lookup['id'] # clobber the name value with the looked-up UUID
+                        except Exception as e:
+                            raise Exception('ERROR: Failed to find exactly one egress Router "{}". Caught exception: {}'.format(egress_router_id, e))
+                    body['model']['edgeRouterHosts'] = [{
+                            "edgeRouterId": egress_router_id,
+                            "serverEgress": server_egress,
+                        }]
+                else:
+                    raise Exception('ERROR: invalid Service model: need only one of binding "endpoints" or hosting "egress_router_id" if "server_host_name" is specified')
+                
             # resolve Edge Router param
-            if edge_router_attributes:
+            if edge_router_attributes and not edge_router_attributes == ['#all']:
                 eprint("WARN: overriding default Service Edge Router Policy #all for new Service {:s}".format(name))
                 body['edgeRouterAttributes'] = edge_router_attributes
+            # TODO: remove when legacy Services API is decommissioned in favor of Platform Services API
+            # results in HTMLv5-compliant URL param singleton with empty string value like ?beta= to invoke the Platform Services API
+            params = {
+                "beta": ''
+            }
 
             response = requests.post(
                 self.session.audience+'core/v2/services',
                 proxies=self.session.proxies,
                 verify=self.session.verify,
                 headers=headers,
-                json=body
+                json=body,
+                params=params
             )
             response_code = response.status_code
         except:
@@ -1179,7 +1300,7 @@ class Network:
                 "findByName": name
             }
             response = requests.get(
-                self.session.audience+'/core/v2/networks',
+                self.session.audience+'core/v2/networks',
                 proxies=self.session.proxies,
                 verify=self.session.verify,
                 headers=headers,
@@ -1219,7 +1340,7 @@ class Network:
                 "authorization": "Bearer " + self.session.token 
             }
             response = requests.get(
-                self.session.audience+'/core/v2/networks/'+network_id,
+                self.session.audience+'core/v2/networks/'+network_id,
                 proxies=self.session.proxies,
                 verify=self.session.verify,
                 headers=headers
@@ -1254,6 +1375,10 @@ class Network:
         :param sleep: SECONDS polling interval
         """
 
+        # use the id of this instance's Network unless another one is specified
+        if type == "network" and not id:
+            id = self.id
+
         now = time.time()
 
         if not wait >= sleep:
@@ -1325,6 +1450,10 @@ class Network:
         :param sleep: SECONDS polling interval
         """
 
+        # use the id of this instance's Network unless another one is specified
+        if type == "network" and not id:
+            id = self.id
+
         now = time.time()
 
         if not wait >= sleep:
@@ -1395,6 +1524,13 @@ class Network:
 
         try:
             headers = { "authorization": "Bearer " + self.session.token }
+
+            params = dict()
+            if not type == "network":
+                params["networkId"] = self.id
+            elif type == "service": 
+                params["beta"] = ''
+
             entity_url = self.session.audience+'core/v2/'
             if type == 'network':
                 entity_url += 'networks/'+self.id
@@ -1402,9 +1538,9 @@ class Network:
                 raise Exception("ERROR: entity UUID must be specified if not a network")
             else:
                 entity_url += type+'s/'+id
-            params = {
-                "networkId": self.id
-            }
+            # TODO: remove "beta" when legacy Services API is decommissioned in favor of Platform Services API
+            # results in HTMLv5-compliant URL param singleton with empty string value like ?beta= to invoke the Platform Services API
+
             response = requests.get(
                 entity_url,
                 proxies=self.session.proxies,
@@ -1445,9 +1581,14 @@ class Network:
         try:
             headers = { "authorization": "Bearer " + self.session.token }
             entity_url = self.session.audience+'core/v2/'+type+'s/'+id
-            params = {
-                "networkId": self.id
-            }
+            # TODO: remove "beta" when legacy Services API is decommissioned in favor of Platform Services API
+            # results in HTMLv5-compliant URL param singleton with empty string value like ?beta= to invoke the Platform Services API
+            params = dict()
+            if not type == "network":
+                params["networkId"] = self.id
+            elif type == "service": 
+                params["beta"] = ''
+
             response = requests.get(
                 entity_url,
                 proxies=self.session.proxies,
@@ -1510,11 +1651,18 @@ class Network:
                 entity_url = self.session.audience+'core/v2/'+type+'s/'+id
                 expect = requests.status_codes.codes.OK
             eprint("WARN: deleting {:s}".format(entity_url))
+            # TODO: remove "beta" when legacy Services API is decommissioned in favor of Platform Services API
+            # results in HTMLv5-compliant URL param singleton with empty string value like ?beta= to invoke the Platform Services API
+            params = dict()
+            if type == "service":
+                params["beta"] = ''
+
             response = requests.delete(
                 entity_url,
                 proxies=self.session.proxies,
                 verify=self.session.verify,
-                headers=headers
+                headers=headers,
+                params=params
             )
             response_code = response.status_code
         except:
@@ -1614,7 +1762,7 @@ RESOURCES = {
     },
     'services': {
         'embedded': "serviceList",
-        'expect': "OK"
+        'expect': "ACCEPTED"
     }
 }
 
