@@ -72,7 +72,7 @@ def main():
         network = netfoundry.Network(session, network_name=network_name)
         network.wait_for_status("PROVISIONED",wait=999,progress=True)
     else:
-        network_id = network_group.create_network(network_name)
+        network_id = network_group.create_network(name=network_name,size="small")['id'] # use "medium" unless demo
         network = netfoundry.Network(session, network_id=network_id)
         network.wait_for_status("PROVISIONED",wait=999,progress=True)
         network = netfoundry.Network(session, network_id=network_id)
@@ -303,7 +303,7 @@ def main():
         customer_router = [er for er in customer_routers if er['name'] == customer_router_name][0]
     # wait for customer router to be PROVISIONED so that registration will be available 
     try:
-        network.wait_for_status("NEW",id=customer_router['id'],type="edge-router",wait=999,progress=True)
+        network.wait_for_status("PROVISIONED",id=customer_router['id'],type="edge-router",wait=999,progress=True)
     except:
         raise
     customer_router_registration = network.get_edge_router_registration(id=customer_router['id'])
@@ -328,7 +328,7 @@ def main():
     print("SUCCESS! The next step is to enroll one or more of your dialer Endpoints on some device(s) and visit one of the demo Service URLs described in the demo document ({doc})."
             "You may also log in to the web console ({nfconsole}) to play with your Network".format(doc="https://developer.netfoundry.io/v2/tools/#demos",nfconsole=network_group.nfconsole))
     for svc in network.services():
-        print("* {name}:\thttp://{url}/".format(name=svc['name'],url=svc['clientHostName']))
+        print("* {name}:\thttp://{url}/".format(name=svc['name'],url=svc['model']['clientIngress']['host']))
 
 if __name__ == '__main__':
     main()
\ No newline at end of file
